/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dirstateparser;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 * @author Tim Boudreau
 */
public class DirstateFile implements Iterable<FileEntry> {
    private final File dirstate;
    private FileChannel channel;
    private ByteBuffer buf;
    private FileInputStream in;
    public DirstateFile(File dirstate) {
        this.dirstate = dirstate;
    }

    private long lastMod;
    public synchronized void open() throws IOException {
        in = new FileInputStream (dirstate);
        lastMod = dirstate.lastModified();
        channel = in.getChannel();
        buf = channel.map(FileChannel.MapMode.READ_ONLY, 0, dirstate.length());
    }

    boolean isFileChanged() {
        return dirstate.lastModified() != lastMod;
    }

    public synchronized void close() throws IOException {
        if (channel != null) {
            channel.close();
            in.close();
            buf = null;
        }
    }

    private synchronized ByteBuffer buf() throws IOException {
        if (isFileChanged()) {
            //XXX may be able to simply extend the mapping of the existing
            //buffer - look into this
            close();
        }
        if (buf == null) {
            open();
        }
        buf.rewind();
        return buf;
    }

    public List<FileEntry> allEntries() throws IOException {
        ByteBuffer workingBuf;
        synchronized (this) {
            ByteBuffer buffer = buf();
            workingBuf = buffer.slice();
        }
        List<FileEntry> entries = new LinkedList<FileEntry>();
        boolean done = false;
        workingBuf.position(40); //Past the 40 byte header
        int pos = 40;
        for (;;) {
            FileEntry e = new FileEntry(pos, this);
            entries.add (e);
            pos += e.length();
            if (pos >= workingBuf.capacity()) {
                break;
            }
        }
        return entries;
    }

    public synchronized ByteBuffer buf(long offset) throws IOException {
        ByteBuffer buffer = buf();
        buffer.position((int) offset); //XXX should remap as needed for > Integer.MAX_VALUE files
        return buffer.slice();
    }

    public byte[] firstParent() throws IOException {
        ByteBuffer b = buf();
        byte[] result = new byte[20];
        b.position(0);
        b.get(result);
        return result;
    }

    public byte[] secondParent() throws IOException {
        ByteBuffer b = buf();
        byte[] result = new byte[40];
        b.position(0);
        b.get(result);
        return result;
    }

    public FileEntry findEntry(String filename) throws IOException {
        int[] pos = positionOf (filename);
        if (pos == null) {
            return null;
        }
        return new FileEntry (pos[0], this);
    }

    private int[] positionOf (String filename) throws IOException {
        ByteBuffer workingBuf;
        synchronized (this) {
            ByteBuffer buffer = buf();
            workingBuf = buffer.slice();
        }
        AsciiCharbuffer seq = new AsciiCharbuffer(workingBuf);
        int flags = Pattern.LITERAL;
        /**
         * if (Utilities.isWindows()) {
         *      flags |= Pattern.CASE_INSENSITIVE;
         * }
         */
        Pattern p = Pattern.compile (filename, flags);
        Matcher m = p.matcher (seq);
        if (m.find()) {
            int pos = m.start() - 17;
            workingBuf.rewind();
            workingBuf.position(pos);
            int end = pos + 17 + filename.length();
            return new int[] { pos, end };
        }
        return null;
    }

    public Iterator<FileEntry> iterator() {
        ByteBuffer buffer;
        try {
            buffer = buf().slice();
            buffer.position(40);
        } catch (IOException ex) {
            Logger.getLogger(DirstateFile.class.getName()).log(Level.SEVERE, null, ex);
            //this will give us a do-nothing iterator
            buffer = ByteBuffer.allocate(0);
        }
        final ByteBuffer b = buf;
        return new Iterator<FileEntry>() {
            int pos;
            public boolean hasNext() {
                return pos < b.capacity();
            }

            public FileEntry next() {
                FileEntry result = new FileEntry (pos, DirstateFile.this);
                try {
                    pos += result.length();
                } catch (IOException ex) {
                    throw new IllegalStateException (ex);
                }
                return result;
            }

            public void remove() {
                throw new UnsupportedOperationException("Not supported yet.");
            }
        };
    }

    private static final class AsciiCharbuffer implements CharSequence {
        private final ByteBuffer buf;

        private AsciiCharbuffer (ByteBuffer buf) {
            this.buf = buf;
        }

        public int length() {
            return buf.limit();
        }

        public char charAt(int index) {
            return (char) buf.get(index);
        }

        public CharSequence subSequence(int start, int end) {
            try {
                buf.rewind();
                buf.position(start);
                buf.limit(end);
                ByteBuffer nue = buf.slice();
                return new AsciiCharbuffer(nue);
            } finally {
                buf.rewind();
                buf.limit(buf.capacity());
            }
        }
    }
}
